#pragma once

#include <cassert>
#include <vector>
#include <stdint.h>
#include <limits>

/// Generation indexing and containers (index, generation)
namespace generational {
    /// Generation element identifier
    struct id {
        using type = uint64_t;
        using idx_t = uint32_t;
        //24bits
        using gen_t = uint32_t;
        using flag_t = uint_fast8_t;

        id(type v): val(v) { }
        id(idx_t index, gen_t generation, flag_t flag = 0): id(merge(index, generation, flag)) { }
        /// Null id
        id(): id(std::numeric_limits<type>::max()) { }
        type val;

        static constexpr size_t TYPE_BITS = sizeof(type) * 8;
        static_assert(TYPE_BITS == 64, "Expect generational::id::type to be 64bit");
        static constexpr size_t HALF_BITS = TYPE_BITS / 2;
        static constexpr type INDEX_MASK = (1l << HALF_BITS) - 1;
        static constexpr size_t INDEX_OFFSET = 0;
        static constexpr type GENERATION_MASK = (1l << 3*HALF_BITS/4) - 1;
        static constexpr size_t GENERATION_OFFSET = HALF_BITS;
        static constexpr type FLAG_MASK = (1l << HALF_BITS/4) - 1;
        static constexpr size_t FLAG_OFFSET = 7*HALF_BITS/4;

        constexpr idx_t index() const { return val & INDEX_MASK; }
        constexpr gen_t generation() const { return (val >> GENERATION_OFFSET) & GENERATION_MASK; }
        constexpr flag_t flag() const { return (val >> FLAG_OFFSET) & FLAG_MASK; }

        static constexpr type merge(idx_t index, gen_t generation = 0, flag_t flag = 0) {
            return (index & INDEX_MASK) | ((generation & GENERATION_MASK) << GENERATION_OFFSET) | ((flag & FLAG_MASK) << FLAG_OFFSET);
        }

        inline bool operator==(const id &i) const { return index() == i.index() && generation() == i.generation(); }
        inline bool operator!=(const id &i) const { return !operator==(i); }
        inline operator bool() const { return index() != id().index(); }
    };

    /// Identifier for target element
    template <typename Target>
    struct key {
        key(id i): val(i) { }
        key(): key(id()) { }
        id val;

        inline bool operator==(const key<Target> &k) const { return val == k.val; }
        inline bool operator!=(const key<Target> &k) const { return val != k.val; }
        inline operator bool() const { return val; }
    };
    /// Index for target element
    template <typename Target>
    struct ref {
        ref(id::idx_t index): val(index) { }
        ref(const key<Target>& k): ref(k.val.index()) { }
        ref(): ref(id().index()) { }
        id::idx_t val;

        inline bool operator==(const ref<Target> &r) const { return val == r.val; }
        inline bool operator!=(const ref<Target> &r) const { return val != r.val; }
        inline operator bool() const { return val != id().index(); }
    };

    /// Closed generational allocator
    class allocator {
    public:
        allocator() { }
        template <typename C>
        allocator(const C& map) {
            for(const auto& [key, _]: map) {
                if (const auto size = entries.size(); key >= size) {
                    entries.resize(key + 1);
                    for (size_t i = size; i < entries.size(); i++) {
                        entries[i].set_live(false);
                        freed.push_back(i);
                    }
                }
                assert(!entries[key].is_alive());
                entries[key].set_live(true);
            }
        }
        id alloc() {
            if(freed.empty()) {
                const auto idx = entries.size();
                entries.emplace_back();
                return id(idx, 0);
            } else {
                const auto idx = freed.back();
                freed.pop_back();
                auto &entry = entries[idx];
                assert(!entry.is_alive());
                entry.set_live(true);
                return id(idx, ++entries[idx]);
            }
        }
        bool is_live(id id) const {
            return id.index() < entries.size() &&
                entries[id.index()].is_alive() &&
                entries[id.index()].generation() == id.generation();
        }
        bool free(id id) {
            if(!is_live(id))
                return false;

            entries[id.index()].set_live(false);
            freed.push_back(id.index());
            return true;
        }

    private:
        struct entry {
            uint_fast32_t val = 0;

            constexpr id::type generation() const { return val & id::GENERATION_MASK; }
            inline id::type operator++() { return (++val) & id::GENERATION_MASK; }
            static constexpr size_t DEAD_OFFSET = id::FLAG_OFFSET - id::GENERATION_OFFSET;
            static constexpr uint_fast32_t DEAD_MASK = 1l << DEAD_OFFSET;
            constexpr bool is_alive() const { return (val & DEAD_MASK) == DEAD_MASK; }
            inline void set_live(bool alive) { val = (val & ~DEAD_MASK) | (static_cast<uint32_t>(!alive) << DEAD_OFFSET); }
        };
        std::vector<entry> entries;
        std::vector<id::type> freed;
    };
}
