#pragma once

#include "./vector.hpp"

namespace generational {

    /// Mostly complete reorder of F to T (bidirectional)
    template<typename F, typename T>
    class heavy_reorder {
    public:
        struct entry: public T {
            entry(ref<F> r, const T &in): T(in), self(r) { }
            template <typename... _Args>
            entry(ref<F> r, _Args &&... __args):
                T(std::forward<_Args>(__args)...), self(r) { }

            ref<F> self;
        };

        std::optional<ref<T>> find(ref<F> r) const {
            if (const auto idx = map.get(r)) {
                return *idx;
            }
            return std::nullopt;
        }
        entry* get(ref<F> r) {
            if (const auto idx = map.get(r)) {
                assert(entries[*idx].self == r);
                return &entries[*idx];
            }
            return nullptr;
        }
        const entry* get(ref<F> r) const {
            if (const auto idx = map.get(r)) {
                assert(entries[*idx].self == r);
                return &entries[*idx];
            }
            return nullptr;
        }
        entry* at(ref<T> r) {
            if (r.val < entries.size()) {
                return &entries[r.val];
            }
            return nullptr;
        }
        const entry* at(ref<T> r) const {
            if (r.val < entries.size()) {
                return &entries[r.val];
            }
            return nullptr;
        }

        template <typename... _Args>
        std::pair<generational::id::idx_t, entry*> emplace_back(ref<F> r, _Args &&... __args) {
            assert(!find(r));
            const generational::id::idx_t idx = entries.size();
            map.set(r, idx);
            auto it = entries.emplace_back(r, std::forward<_Args>(__args)...);
            return std::make_pair(idx, &it);
        }
        /// Place at idx moving childs if needed
        template <typename... _Args>
        entry& set_emplace(ref<F> r, generational::id::idx_t at, _Args &&... __args) {
            assert(!find(r) && at <= entries.size());
            map.set(r, at);
            auto it = entries.emplace(entries.begin()+at, r, std::forward<_Args>(__args)...);
            for (auto next = it+1; next != entries.end(); ++next) {
                auto& cell = map.at(next->self.val);
                assert(cell.has_value());
                ++cell.value();
            }
            return *it;
        }

        void erase(ref<F> r) {
            if (const auto idx = find(r)) {
                map.erase(r);
                auto it = entries.erase(entries.begin() + idx.value().val);
                for (; it != entries.end(); ++it) {
                    auto& cell = map.at(it->self.val);
                    assert(cell.has_value());
                    --cell.value();
                }
            }
        }
        void erase(ref<F> from, size_t count) {
            if (const auto idx = find(from)) {
                auto it = entries.begin() + idx.value().val;
                const auto start = it;
                const auto last = it + count;
                for (; it != last; ++it)
                    map.erase(it->self.val);
                for (; it != entries.end(); ++it) {
                    auto& cell = map.at(it->self.val);
                    assert(cell.has_value());
                    cell.value() -= count;
                }
                entries.erase(start, last);
            }
        }

        template<typename apply>
        void iter(apply fn) const {
            for (const auto& entry: entries) {
                fn(entry);
            }
        }
        template<typename apply>
        void for_each(apply fn) {
            for (auto& entry: entries) {
                fn(entry);
            }
        }

        size_t size() const { return entries.size(); }
        void reserve(size_t n) {
            map.reserve(n);
            entries.reserve(n);
        }

    private:
        std::vector<entry> entries;
        view_vector<F, generational::id::idx_t> map;
    };

    template<typename F, typename T>
    struct hierarchy_entry: public T {
        using offset = generational::id::idx_t;
        hierarchy_entry(const key<F> &p, offset n, const T &in) : T(in), parent(p), childs(n) {}
        template <typename... _Args>
        hierarchy_entry(const key<F>& p, offset n, _Args &&... __args):
            T(std::forward<_Args>(__args)...), parent(p), childs(n) { }

        key<F> parent;
        offset childs;
    };

    /// Sorted hierarchy with parent and next child idx
    template<typename F, typename T>
    class hierarchy: public heavy_reorder<F, hierarchy_entry<F, T>> {
    public:
        using parent = heavy_reorder<F, hierarchy_entry<F, T>>;
        using entry = typename parent::entry;

        /// Add single root node
        template <typename... _Args>
        std::pair<generational::id::idx_t, entry*> emplace_back(ref<F> r, _Args &&... __args) {
            return parent::emplace_back(r, generational::id(), 0, std::forward<_Args>(__args)...);
        }
        /// Add node with parent
        template <typename... _Args>
        entry& set_emplace(ref<F> r, const key<F>& parent, _Args &&... __args) {
            assert(!parent::find(r));
            if (const auto pid = parent::find(parent.val.index())) {
                const ref<hierarchy_entry<F, T>> cid = pid.value().val+1;
                auto &it = parent::set_emplace(r, cid, parent, 0, std::forward<_Args>(__args)...);
                move_childs(it.parent, 1);
                return it;
            } else {
                assert(!parent);
                return *emplace_back(r, std::forward<_Args>(__args)...).second;
            }
        }

        template<typename fn>
        void erase(ref<F> r, const fn& childApply) {
            if (const auto idx = parent::find(r)) {
                auto self = parent::at(idx.value());
                const auto parent = self->parent;
                const auto childs = self->childs;
                parent::erase(r);
                move_childs(parent, -1);
                for (size_t i = 0; i < childs; i++) {
                    auto child = parent::at(idx.value().val + i);
                    child->parent = parent;
                    childApply(child);
                }
            }
        }
        template<typename fn>
        void erase_recursive(ref<F> r, const fn& apply) {
            if (const auto idx = parent::find(r)) {
                auto self = parent::at(idx.value());
                const auto parent = self->parent;
                const auto to_delete = self->childs+1;
                for (size_t i = 0; i < to_delete; i++) {
                    apply(parent::at(idx.value().val + i));
                }
                parent::erase(r, to_delete);
                move_childs(parent, -to_delete);
            }
        }

    private:
        void move_childs(const key<F>& from, int of) {
            ref<F> cur = from.val.index();
            while(cur) {
                auto it = parent::get(cur);
                assert(it);
                it->childs = of + it->childs;
                cur = it->parent.val.index();
            }
        }
    };

}