#include "Universe.hpp"

#include "Elements.hpp"

using namespace world;

inline bool intersectRay(const geometry::Ray& ray, const Node& node, const Elements& l) {
    assert(node.content);
    // MAYBE: check obb
    return ray.intersect(aabb::From(node.content->getBoundingBox(l), node.absolute.rotation) + node.absolute.position)
        != lfaabb::ContainmentType::Disjoint;
}
inline bool containPoint(const world_pos& point, const Node& node, const Elements& l) {
    assert(node.content);
    // MAYBE: check obb
    return (aabb::From(node.content->getBoundingBox(l), node.absolute.rotation) + node.absolute.position).contains(point);
}

template<typename Tr = Universe::voxel_query_target, typename Vp = node_voxel_pos, typename CC, typename Rs = Universe::query_result>
inline void raycastVoxel(const geometry::Ray& ray, bool solid_only, const node_id& id, const transform& absolute, const CC& chunks, Rs& target, float& targetDist) {
    if (chunks.empty())
        return;

    const auto inv_rot = glm::conjugate(absolute.rotation);
    const auto localRay = geometry::Ray((lpivot)inv_rot * (ray.from - absolute.position), inv_rot * ray.dir, targetDist);
    std::shared_ptr<world::EdittableChunk> chunk = nullptr;
    chunk_pos chunk_vec(INT_MAX);
    geometry::Ray::iterator it(localRay);
    glm::llvec3 pos;
    const auto setTargetDist = [&] {
        const auto pDist = glm::length((glm::lfvec3)pos - localRay.from);
        if (pDist < targetDist) {
            targetDist = pDist;
            return true;
        }
        return false;
    };
    // MAYBE: fastforward with chunk iterator
    while (it.next(pos)) {
        const chunk_pos cPos = glm::divide(pos);
        if (cPos != chunk_vec) {
            if (chunks.inRange(cPos)) {
                if (chunk = chunks.findInRange(cPos); chunk) {
                    chunk_vec = cPos;
                } else {
                    if (setTargetDist())
                        target = Universe::area_query_out_of_range{id.val, cPos};

                    break;
                }
            } else {
                chunk = nullptr;
            }
        }
        if (chunk != nullptr) {
            const auto voxel = chunk->getAt(glm::modulo(pos));
            if (voxel.is_material() && (!solid_only || voxel.is_solid())) {
                if (setTargetDist())
                    target = Tr(Vp{id.val, pos}, voxel);

                break;
            }
        }
    }
}

template<typename CC>
inline std::optional<Universe::voxel_query_target> voxelAt(const world_pos& point, bool solid_only, const node_id& id, const transform& absolute, const CC& chunks) {
    if (chunks.empty())
        return std::nullopt;

    const auto localPoint = (lpivot)glm::conjugate(absolute.rotation) * (point - absolute.position);
    const chunk_pos cPos = glm::divide(localPoint);
    if (const auto chunk = chunks.findInRange(cPos)) {
        const auto voxel = chunk->getAt(glm::modulo(localPoint));
        if (voxel.is_material() && (!solid_only || voxel.is_solid())) {
            return Universe::voxel_query_target(node_voxel_pos{id, localPoint}, voxel);
        }
    }
    return std::nullopt;
}

Universe::query_result Universe::Raycast(const Elements& l, const geometry::Ray &in, bool solid_only, const node_id& ignore) {
    geometry::Ray ray = in;
    query_result target;
    // TODO: use bounding hierarchy
    l.nodes.iter([&](const node_id &id, const Node &node) {
        const auto type = Elements::GetType(id);
        if (id == ignore || type == Elements::Type::None || node.content == nullptr || !intersectRay(ray, node, l))
            return;

        switch (type) {
            case Elements::Type::Area:
                raycastVoxel(ray, solid_only, id, node.absolute, NodeOf<Area>::Make(&node)->get()->getChunks(), target, ray.dist);
                break;

            case Elements::Type::Part:
                raycastVoxel(ray, solid_only, id, node.absolute, *NodeOf<Part>::Make(&node)->get(), target, ray.dist);
                break;

            default: {
                // TODO: get intersect point OBB ray -> point
                target = element_query_target{id, world_pos(0)};
                const auto pDist = glm::length(node.absolute.position - ray.from);
                //FIXME: include bounding assert(pDist <= ray.dist);
                if (pDist < ray.dist)
                    ray.dist = pDist;
                break;
            }
        }
    });
    return target;
}

Universe::element_query_result Universe::Raycast(const Elements& l, const geometry::Ray &in, const node_id& ignore) {
    const auto result = Raycast(l, in, false, ignore);
    if (const auto voxel = std::get_if<voxel_query_target>(&result)) {
        return element_query_target{voxel->pos.node, voxel->pos.voxel};
    } else if (const auto element = std::get_if<element_query_target>(&result)) {
        return *element;
    } else {
        return std::nullopt;
    }
}

Universe::area_query_result Universe::RaycastArea(const Elements& l, const geometry::Ray& in, bool solid_only) {
    geometry::Ray ray = in;
    area_query_result target;
    for (const auto& ref: l.areas) {
        const auto id = l.withFlag(ref);
        const auto node = l.findArea(id.val);
        assert(node);
        if (!intersectRay(ray, *node, l))
            continue;

        raycastVoxel<area_query_target, area_voxel_pos>(ray, solid_only, id, node->absolute, node->get()->getChunks(), target, ray.dist);
    }
    return target;
}

Universe::query_result Universe::At(const Elements& l, const world_pos &point, bool solid_only, const node_id& ignore) {
    query_result target;
    // TODO: use bounding hierarchy
    l.nodes.any([&](const node_id &id, const Node &node) {
        const auto type = Elements::GetType(id);
        if (id == ignore || type == Elements::Type::None || node.content == nullptr || !containPoint(point, node, l))
            return false;

        switch (type) {
            case Elements::Type::Area: {
                if (const auto result = voxelAt(point, solid_only, id, node.absolute, NodeOf<Area>::Make(&node)->get()->getChunks())) {
                    target = result.value();
                    return true;
                }
                return false;
            }
            case Elements::Type::Part: {
                if (const auto result = voxelAt(point, solid_only, id, node.absolute, *NodeOf<Part>::Make(&node)->get())) {
                    target = result.value();
                    return true;
                }
                return false;
            }
            default: {
                target = element_query_target{id, (lpivot)glm::conjugate(node.absolute.rotation) * (point - node.absolute.position)};
                return true;
            }
        }
    });
    return target;
}

Universe::element_query_result Universe::At(const Elements& l, const world_pos &point, const node_id& ignore) {
    const auto result = At(l, point, false, ignore);
    if (const auto voxel = std::get_if<voxel_query_target>(&result)) {
        return element_query_target{voxel->pos.node, voxel->pos.voxel};
    } else if (const auto element = std::get_if<element_query_target>(&result)) {
        return *element;
    } else {
        return std::nullopt;
    }
}

bool Universe::IsRangeFree(const Elements& l, const node_voxel_ref &center_ref, const geometry::Volume& volume) {
    //TODO: handle transform (needs OBB-OBB intersect)
    const auto center = l.computeAbsolute(center_ref.node, center_ref.voxel);

    //FIXME: use hierarchy bounding box
    return !l.nodes.any([&](const node_id& id, const Node& node) {
        return id.val.index() != center_ref.node.val && volume.intersects(center,
            obb(node.content->getBoundingBox(l), node.absolute.rotation) + node.absolute.position);
    });
}

/*
bool Universe::move(element_pos &pos, const glm::vec3 &vel, int density, float radius) const {
    const auto dir = glm::normalize(vel);
    const auto velocity = vel * glm::vec3(density);
    const auto from = pos * density + dir;
    const auto result = raycast(Ray(from, dir, glm::length(velocity) + radius));
    if (auto target = std::get_if<ray_target>(&result)) {
        const auto target_dist = from.dist(element_pos(target->offset + target->pos.second, density)) - radius;
        pos += vel * glm::vec3(target_dist / glm::length(vel));
        return true;
    } else if(auto out = std::get_if<ray_out_of_range>(&result)) {
        const auto out_dist = from.dist(element_pos(out->offset + glm::multiply(out->pos.second), density)) - radius - CHUNK_LENGTH;
        pos += vel * glm::vec3(out_dist / glm::length(vel));
        return true;
    }

    pos += vel;
    return false;
}
bool Universe::collide(const element_pos &pos, const glm::vec3 &vel, int density, float radius) const {
    const auto dir = glm::normalize(vel);
    const auto velocity = vel * glm::vec3(density);
    const auto from = pos * density + dir;
    return !std::holds_alternative<std::monostate>(raycast(Ray(from, dir, glm::length(velocity) + radius)));
}
*/